// FastArray.h

#pragma once

#ifndef _FAST_ARRAY_H_
#define _FAST_ARRAY_H_

//#include "allocator_extern.h"
#include "allocator_id.h"


#include <vector>


//--------------------------------------------------------------
// FastArray
//
// warning: this array class move elements using memcpy when
// growing
//--------------------------------------------------------------

template<typename T>
class FastArray
{
public:
	//typedef typename std::vector<T>::iterator iterator;
	//typedef typename std::vector<T>::const_iterator const_iterator;

	class const_iterator
	{
		//friend void _Advance<class _InIt, class _Diff> ( _InIt & it, _Diff distance, fast_deque_iterator_tag );

	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
			typedef std::random_access_iterator_tag iterator_category;
			typedef const T value_type;
			typedef ptrdiff_t difference_type;
			typedef ptrdiff_t distance_type;	// retained
			typedef const T * pointer;
			typedef const T & reference;
		//};

		friend FastArray;

	public:
		const_iterator ( const T * current )
			: current_ ( current )
		{}

		const_iterator()
			: current_ ( 0 )
		{}

		const T & operator*() const
		{
			return * current_;
		}

		const T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const const_iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const const_iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		const_iterator & operator++ () // preincrement
		{
			++current_;
			return *this;
		}

		const_iterator operator++ ( int ) // postincrement
		{
			const_iterator retValue = *this;
			++current_;
			return retValue;
		}

		const_iterator & operator-- () // predecrement
		{
			--current_;
			return *current_;
		}

		const_iterator operator-- ( int ) // postdecrement
		{
			const_iterator retValue = *this;
			--*this;
			return retValue;
		}

	//private:
	public:
		const T * current_;
	};

	class iterator
	{
		//friend void _Advance<class _InIt, class _Diff> ( _InIt & it, _Diff distance, fast_deque_iterator_tag );

	public:
		//template<class _Ty>
		//struct iterator_traits//<_Ty *>
		//{	// get traits from pointer
			typedef std::random_access_iterator_tag iterator_category;
			typedef T value_type;
			typedef ptrdiff_t difference_type;
			typedef ptrdiff_t distance_type;	// retained
			typedef T * pointer;
			typedef T & reference;
		//};

		friend FastArray;

	public:
		iterator ( T * current )
			: current_ ( current )
		{}

		iterator()
			: current_ ( 0 )
		{}

		operator const_iterator() const
		{
			return const_iterator ( current_ );
		}

		T & operator*() const
		{
			return * current_;
		}

		T * operator->() const
		{
			return &(*current_);
		}

		bool operator== ( const iterator & rhs ) const
		{
			return current_ == rhs.current_;
		}

		bool operator!= ( const iterator & rhs ) const
		{
			return current_ != rhs.current_;
		}

		iterator & operator++ () // preincrement
		{
			++current_;
			return *this;
		}

		iterator operator++ ( int ) // postincrement
		{
			iterator retValue = *this;
			++current_;
			return retValue;
		}

		iterator & operator-- () // predecrement
		{
			--current_;
			return *current_;
		}

		iterator operator-- ( int ) // postdecrement
		{
			iterator retValue = *this;
			--*this;
			return retValue;
		}

	//private:
	public:
		T * current_;
	};

	explicit FastArray()
		: first_ ( 0 )
		, last_ ( 0 )
		, end_ ( 0 )
	{}

	void clear()
	{
		T * it_begin = first_;
		T * it_last = last_;
		if ( it_begin != it_last )
		{
			for ( T * it = it_begin; it != it_last; ++it )
			{
				it->~T(); // call destructor
			}
			deallocate ( (byte*)first_, (byte*)end_ );
		}
		first_ = 0;
		last_ = 0;
		end_ = 0;
	}

	~FastArray()
	{
		clear();
	}

private:
	byte * allocate ( size_t size )
	{
		void * address;
		if ( size <= 16 )
			address = FixedSmallObjAllocator<16,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 32 )
			address = FixedSmallObjAllocator<32,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 64 )
			address = FixedSmallObjAllocator<64,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 128 )
			address = FixedSmallObjAllocator<128,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 256 )
			address = FixedSmallObjAllocator<256,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else if ( size <= 512 )
			address = FixedSmallObjAllocator<512,VariableSizeAllocatorId,OwnPageAllocator>::allocate();
		else
			address = new byte[size];
		return (byte*)address;
	}

	void deallocate ( byte * first, byte * end )
	{
		if ( first != end )
		{
			size_t size = end - first;
			if ( size <= 16 )
				FixedSmallObjAllocator<16,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else if ( size <= 32 )
				FixedSmallObjAllocator<32,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else if ( size <= 64 )
				FixedSmallObjAllocator<64,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else if ( size <= 128 )
				FixedSmallObjAllocator<128,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else if ( size <= 256 )
				FixedSmallObjAllocator<256,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else if ( size <= 512 )
				FixedSmallObjAllocator<512,VariableSizeAllocatorId,OwnPageAllocator>::deallocate ( first );
			else
				delete[] first;
		}
	}

public:
	iterator begin()
	{
		return iterator ( first_ );
	}

	const_iterator begin() const
	{
		return const_iterator ( first_ );
	}

	iterator end()
	{
		return iterator ( last_ );
	}

	const_iterator end() const
	{
		return const_iterator ( last_ );
	}

	void push_back ( T value )
	{
//#ifdef _DEBUG
//		unsigned debug_size_before = last_ - first_;
//#endif // _DEBUG

		if ( last_ == end_ )
		{
			if ( first_ )
			{
				size_t size = (byte*)end_ - (byte*)first_;
				size_t new_size = 2 * size;
				byte * new_first = allocate ( new_size );
				memcpy ( new_first, first_, size );
				deallocate ( (byte*) first_, (byte*) end_ );
				first_ = (T*)new_first;
				end_ = (T*) ( new_first + new_size );
				T * current = (T*) ( new_first + size );
				new ( current ) T ( value );
				++current;
				last_ = current;
			}
			else
			{
				size_t size = start_chunk_*sizeof(T);
				byte * first = allocate ( size );
				first_ = (T*)first;
				end_ = (T*) ( first + size );
				T * current = (T*) first;
				new ( current ) T ( value );
				++current;
				last_ = (T*) current;
			}
		}
		else
		{
			new ( last_ ) T ( value );
			++last_;
		}
		
//#ifdef _DEBUG
//		unsigned debug_size_after = last_ - first_;
//		XASSERT ( debug_size_before + 1 == debug_size_after );
//#endif // _DEBUG

	}

	bool empty() const
	{
		return first_ == last_;
	}

	size_t size() const
	{
		return last_ - first_;
	}

	T& operator[] ( unsigned index )
	{
		return first_[index];
	}

private:
	static const unsigned start_chunk_ = 4;
	T * first_; // pointer to beginning of array
	T * last_; // pointer to current end of sequence
	T * end_; // pointer to end of array
};

//--------------------------------------------------------------

#endif // _FAST_ARRAY_H_
