// smartpointer.h

#ifndef _SMARTPOINTER_H_
#define _SMARTPOINTER_H_

//------------------------------------------------------------------------------
// DeleteObjectStrategy
//------------------------------------------------------------------------------

template < typename T >
class CppDelete
{
public:
	static inline void delete_object ( T * obj )
	{
		delete obj;
	}
};

//------------------------------------------------------------------------------
// RefCountBasePtr
//------------------------------------------------------------------------------

template < typename T, class DeleteObjectStrategy = CppDelete < T > >
class RefCountBasePtr
{
public:
	RefCountBasePtr() : obj_(0)
	{}
	
	RefCountBasePtr ( T * obj ) : obj_(obj)
	{
		Lock();
	}
	
	RefCountBasePtr ( const RefCountBasePtr & rhs ) : obj_ ( rhs.obj_ )
	{
		Lock();
	}
	
	RefCountBasePtr & operator= ( const RefCountBasePtr & rhs )
	{
		if ( this == & rhs )
			return *this;

		Unlock();

		obj_ = rhs.obj_;

		Lock();

		return *this;
	}
	
	~RefCountBasePtr()
	{
		Unlock();
	}

	T * operator->()
	{
		return obj_;
	}

	const T * operator-> () const
	{
		return obj_;
	}

	//operator T * ()
	//{
	//	return obj_;
	//}

	//operator const T * () const
	//{
	//	return obj_;
	//}

	operator bool () const
	{
		return obj_ != 0;
	}

	T * get()
	{
		return obj_;
	}

	const T * get() const
	{
		return obj_;
	}

	// hierarchy conversion ([3rd, 13.6.3.1])

	template<typename T2>
	operator RefCountBasePtr<T2>()
	{
		return RefCountBasePtr<T2>(obj_);
	}

private:

	void inline Lock()
	{
		if ( obj_ )
			++((obj_)->RefCountBase::ref_counter_);
	}

	void inline Unlock()
	{
		if ( obj_ && ( ( --(obj_)->RefCountBase::ref_counter_) == 0 ) )
			DeleteObjectStrategy::delete_object ( obj_ );
	}

private:
	T * obj_;
};

//------------------------------------------------------------------------------
// RefCountBaseManualPtr
//------------------------------------------------------------------------------

template < typename T, class DeleteObjectStrategy = CppDelete < T > >
class RefCountBaseManualPtr
{
public:
	//RefCountBaseManualPtr() : obj_(0)
	//{}
	
	//RefCountBaseManualPtr ( T * obj ) : obj_(obj)
	//{
	//	Lock();
	//}
	
	//RefCountBaseManualPtr ( const RefCountBaseManualPtr & rhs ) : obj_ ( rhs.obj_ )
	//{
	//	Lock();
	//}
	//
	//RefCountBaseManualPtr & operator= ( const RefCountBaseManualPtr & rhs )
	//{
	//	if ( this == & rhs )
	//		return *this;

	//	Unlock();

	//	obj_ = rhs.obj_;

	//	Lock();

	//	return *this;
	//}
	
	//~RefCountBaseManualPtr()
	//{
	//	Unlock();
	//}

	T * operator->()
	{
		return obj_;
	}

	const T * operator-> () const
	{
		return obj_;
	}

	//operator T * ()
	//{
	//	return obj_;
	//}

	//operator const T * () const
	//{
	//	return obj_;
	//}

	operator bool () const
	{
		return obj_ != 0;
	}

	T * get()
	{
		return obj_;
	}

	const T * get() const
	{
		return obj_;
	}

	// hierarchy conversion ([3rd, 13.6.3.1])

	template<typename T2>
	operator RefCountBaseManualPtr<T2>()
	{
		return RefCountBaseManualPtr<T2>(obj_);
	}

public:

	void inline Lock()
	{
		ASSERT ( !obj_ || (obj_)->RefCountBase::ref_counter_ >= 0 );
		if ( obj_ )
			++((obj_)->RefCountBase::ref_counter_);
	}

	void inline Unlock()
	{
		ASSERT ( !obj_ || (obj_)->RefCountBase::ref_counter_ >= 1 );
		if ( obj_ && ( ( --(obj_)->RefCountBase::ref_counter_) == 0 ) )
			DeleteObjectStrategy::delete_object ( obj_ );
	}

public:
	T * obj_;
};

//------------------------------------------------------------------------------

#endif // _SMARTPOINTER_H_